(* file: map.ml
   author: Bob Muller

   CS1103 Computer Science I Honors

   Lecture Notes for Meeting 1 of Week 6

   Topics:
     1. Two Small Items
        A. operators are functions
        B. record "update" notation
     2. Modelling the "state" of your application
        A. handling mouse events
     3. List.map & Friends

  1. Two Small Items ***********************************************

    A. operators are functions

    You can use the built-in operators as functions by enclosing them
    in parentheses:

    # let a = (+) 2 3;;
    val a : int = 5

    NB: The *, *. and ** operators require spaces on the sides to avoid
    confusion with comments.

    # let a = ( * ) 2 3;;
    val a : int = 6

    You can define an operator using the parenthesized notation:

    # let (%) m n = m mod n;;
    val ( % ) : int -> int -> int = <fun>

    B. Notation for record "update"

    You can define a new record by using an old one.

    # type point = {x : int; y : int; z : int};;
    type point = {x : int; y : int; z : int};;

    # let p = {x = 1; y = 2; z = 3};;
    val p : point = {x = 1; y = 2; z = 3}

    # let q = { p with z = 4 };;
    val q : point = {x = 1; y = 2; z = 4}


   2. Modelling the State of an application ******************************

   Applications that model some sort of system (real or imagined) usually
   use a combination of values of various kinds to model the state of the 
   system being modelled.

   In the setting sun example, the "state" of the system, i.e., the thing
   that was changing over time, was the y-coordinate of the sun --- it
   increased as the sun set.  So we had:

   type model = float

   Real systems are far more complicated. For the sun example, it might be
   down, up, rising or setting. So our model can be enhanced:

   type state = Down | Up | Rising | Setting
   type model = {state : state; y : float}

   Now, if the state is Setting, our update function can increase y on 
   each clock-tick but if the state is Rising, our code will decrease y 
   on each clock tick.  (See the full applicationin the sun/ folder).

(* update : model -> model World.t 
*)
let update ({state; y} as model : model) =
  match state with
  | Setting -> let newY = min displayHeight (y +. dropAmount)
               in
               World { model with y = newY}

  | Rising -> let newY = max 0.0 (y -. dropAmount)
              in
              World { model with y = newY}

  | _ -> World model

(* handleMouse : model -> float -> float -> event -> World.t
*)
let handleMouse ({state; y} as model) x y event =
  match (event, state) with
  | ("button_down", Up)      -> World { model with state = Setting}
  | ("button_down", Setting) -> World { model with state = Rising}
  | ("button_down", Down)    -> World { model with state = Rising}
  | ("button_down", Rising)  -> World { model with state = Setting}
  | _                        -> World model

 *)

(* 3. List.map and friends.  *********************************************

  double : int -> int
*)
let double n = n * 2

(* doubleAll : int list -> int list
*)
let rec doubleAll ns =
  match ns with
  | [] -> []
  | n :: ns -> double n :: doubleAll ns

(* lengthAll : ('a list) list -> int list
 *)
let rec lengthAll xxs =
  match xxs with
  | [] -> []
  | xs :: xxs -> List.length xs :: lengthAll xxs

(* These functions are essentially identical, they both build up a
   result list from an input list by applying a function to each 
   element of the input list. They vary in that one is applying the 
   double function while the other is applying the List.length function.

   We can abstract with respect to this variation by replacing the varying
   part with a variable (!):

   map : ('a -> 'b) -> 'a list -> 'b list
*)
let rec map f xs =
  match xs with
  | [] -> []
  | x :: xs -> f x :: map f xs

(* We can now define our doubleAll and lengthAll functions as:
*)
let doubleAll ns = map double ns
let lengthAll xxs = map List.length xxs

(* As we'll see, we can (and it would be better to) define these 
   two functions as:
*)
let doubleAll = map double
let lengthAll = map List.length
                    
